import dataclasses
from unittest import TestCase

from persisty.errors import PersistyError
from persisty.impl.mem.mem_store import MemStore
from persisty.impl.mem.mem_store_factory import MemStoreFactory
from persisty.search_filter.filter_factory import filter_factory
from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.store.store_abc import StoreABC
from persisty.store_meta import get_meta, StoreMeta
from tests.fixtures.number_name import NumberName, NUMBER_NAMES
from tests.fixtures.storage_tst_abc import StoreTstABC
from tests.fixtures.super_bowl_results import (
    SuperBowlResult,
    SUPER_BOWL_RESULTS,
)


class TestMemStore(TestCase, StoreTstABC):
    def new_super_bowl_results_store(self) -> StoreABC:
        factory = MemStoreFactory(
            get_meta(SuperBowlResult),
            {r.code: dataclasses.replace(r) for r in SUPER_BOWL_RESULTS},
        )
        return factory.create()

    def new_number_name_store(self) -> StoreABC:
        # noinspection PyTypeChecker
        factory = MemStoreFactory(
            get_meta(NumberName),
            {str(r.id): dataclasses.replace(r) for r in NUMBER_NAMES},
        )
        return factory.create()

    def test_search_all_sorted(self):
        store = self.new_super_bowl_results_store()
        filters = filter_factory(SuperBowlResult)
        self.assertEqual(
            list(reversed(SUPER_BOWL_RESULTS)),
            list(store.search_all(INCLUDE_ALL, filters.year.desc())),
        )
        self.assertEqual(
            list(reversed(SUPER_BOWL_RESULTS[17:37])),
            list(
                store.search_all(
                    filters.year.gte(1984) & filters.year.lt(2004), filters.year.desc()
                )
            ),
        )

    def test_mem_store_no_dict(self):
        store = MemStoreFactory(get_meta(NumberName)).create()
        created = store.create(NumberName(value=1, title="One"))
        loaded = store.read(str(created.id))
        self.assertEqual(loaded, created)

    def test_mem_store_delete_missing_key(self):
        store = MemStore(get_meta(NumberName))
        self.assertFalse(store.delete("missing_key"))

    def test_mem_store_create_missing_key(self):
        store_meta = get_meta(NumberName)
        store_meta = StoreMeta(
            name=store_meta.name,
            attrs=tuple(
                dataclasses.replace(
                    f,
                    create_generator=None,
                    update_generator=None,
                    creatable=True,
                    updatable=True,
                )
                for f in store_meta.attrs
            ),
        )
        store = MemStore(store_meta)
        with self.assertRaises(PersistyError):
            store.create(NumberName())
        self.assertEqual(0, store.count())

    def test_mem_store_update_no_key(self):
        store_meta = get_meta(NumberName)
        store_meta = StoreMeta(
            name=store_meta.name,
            attrs=tuple(
                dataclasses.replace(
                    f,
                    create_generator=None,
                    update_generator=None,
                    creatable=True,
                    updatable=True,
                )
                for f in store_meta.attrs
            ),
        )
        store = MemStore(store_meta)
        with self.assertRaises(PersistyError):
            store.update(NumberName(title="foobar"))
        self.assertEqual(0, store.count())



from typing import Optional
from unittest import TestCase
from uuid import UUID

from persisty.factory.default_store_factory import DefaultStoreFactory
from persisty.impl.mem.mem_store import MemStore
from persisty.store_meta import get_meta
from persisty.stored import stored


@stored
class Contact:
    id: UUID
    name: str
    email: Optional[str]
    phone: Optional[str]


class TestDefaultStoreFactory(TestCase):
    def test_getters(self):
        meta = get_meta(Contact)
        store = MemStore(meta)
        factory = DefaultStoreFactory(store)
        self.assertEqual(meta, factory.get_meta())
        self.assertEqual(store, factory.create(None))



from unittest import TestCase

from persisty.util.encrypt_at_rest import encrypt, decrypt


class TestEncryptAtRest(TestCase):
    def test_round_trip(self):
        msg = "This is a test"
        encrypted = encrypt(msg)
        self.assertNotEqual(encrypted, msg)
        decrypted = decrypt(encrypted)
        self.assertEqual(decrypted, msg)

    def test_invalid(self):
        encrypted = None
        # noinspection PyBroadException
        try:
            msg = "This is a test"
            encrypted = decrypt(msg)
        except:
            pass
        self.assertIsNone(encrypted)



from dataclasses import dataclass
from unittest import TestCase
from uuid import uuid4

from boto3.dynamodb.conditions import Attr as DynAttr

from persisty.attr.attr_filter import AttrFilter
from persisty.attr.attr_filter_op import AttrFilterOp
from persisty.util import UNDEFINED


@dataclass
class FooBar:
    foo: int
    bar: str


class TestAttrFilter(TestCase):
    def test_match(self):
        attr_filter = AttrFilter("foo", AttrFilterOp.eq, 10)
        self.assertTrue(attr_filter.match(FooBar(10, "bar")))
        self.assertFalse(attr_filter.match(FooBar(11, "bar")))

    def test_match_starts_with(self):
        attr_filter = AttrFilter("bar", AttrFilterOp.startswith, "zap")
        self.assertTrue(attr_filter.match(FooBar(10, "zapbang")))
        self.assertFalse(attr_filter.match(FooBar(10, "zopbang")))

    def test_match_type_error(self):
        attr_filter = AttrFilter("bar", AttrFilterOp.lt, "zap")
        self.assertFalse(attr_filter.match(FooBar(10, uuid4())))

    def test_build_filter_expression_starts_with(self):
        attr_filter = AttrFilter("foo", AttrFilterOp.startswith, "bar")
        filter_expression, handled = attr_filter.build_filter_expression(tuple())
        self.assertTrue(handled)
        self.assertEqual(DynAttr("foo").begins_with("bar"), filter_expression)
        attr_filter = AttrFilter("foo", AttrFilterOp.startswith, 10)
        filter_expression, handled = attr_filter.build_filter_expression(tuple())
        self.assertTrue(handled)
        self.assertEqual(DynAttr("foo").begins_with(10), filter_expression)

    def test_build_filter_expression_exists(self):
        attr_filter = AttrFilter("foo", AttrFilterOp.exists, UNDEFINED)
        filter_expression, handled = attr_filter.build_filter_expression(tuple())
        self.assertTrue(handled)
        self.assertEqual(DynAttr("foo").exists(), filter_expression)

    def test_build_filter_expression_not_exists(self):
        attr_filter = AttrFilter("foo", AttrFilterOp.not_exists, UNDEFINED)
        filter_expression, handled = attr_filter.build_filter_expression(tuple())
        self.assertTrue(handled)
        self.assertEqual(DynAttr("foo").not_exists(), filter_expression)

    def test_build_filter_expression_oneof(self):
        attr_filter = AttrFilter("bar", AttrFilterOp.oneof, ["zap", "bang"])
        filter_expression, handled = attr_filter.build_filter_expression(tuple())
        self.assertFalse(handled)
        self.assertIsNone(filter_expression)



from typing import Union
from unittest import TestCase

from dataclasses import dataclass

from persisty.util import to_base64, from_base64, secure_hash, dataclass_to_params
from persisty.util.undefined import Undefined, UNDEFINED


class TestUtil(TestCase):
    def test_base64_round_trip(self):
        msg = "This is a test"
        b64 = to_base64(msg)
        self.assertNotEqual(b64, msg)
        decoded = from_base64(b64)
        self.assertEqual(decoded, msg)

    def test_base64_round_trip_dict(self):
        payload = dict(msg="This is a test")
        b64 = to_base64(payload)
        decoded = from_base64(b64)
        self.assertEqual(decoded, payload)

    def test_base64_round_trip_none(self):
        payload = None
        b64 = to_base64(payload)
        decoded = from_base64(b64)
        self.assertEqual(decoded, payload)

    def test_secure_hash(self):
        payload = dict(msg="This is a test")
        hash = secure_hash(payload)
        payload["msg"] += "!"
        updated_hash = secure_hash(payload)
        self.assertNotEqual(hash, updated_hash)

    def test_dataclass_to_params(self):
        @dataclass
        class Foo:
            title: Union[str, Undefined, type(None)] = UNDEFINED

        self.assertEqual({}, dataclass_to_params(Foo()))
        self.assertEqual(dict(title="Foobar"), dataclass_to_params(Foo("Foobar")))



from abc import ABC, abstractmethod
from copy import deepcopy
from datetime import datetime, timezone
from typing import Tuple

from dataclasses import dataclass, replace
from uuid import uuid4, UUID

import marshy
from dateutil.relativedelta import relativedelta

from persisty.attr.attr import Attr
from persisty.attr.attr_filter import AttrFilter
from persisty.attr.attr_filter_op import AttrFilterOp
from persisty.batch_edit import BatchEdit
from persisty.errors import PersistyError
from persisty.result_set import ResultSet
from persisty.search_filter.filter_factory import filter_factory
from persisty.search_filter.exclude_all import EXCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.search_order.search_order import SearchOrder
from persisty.search_order.search_order_attr import SearchOrderAttr
from persisty.store.store_abc import StoreABC, T
from tests.fixtures.super_bowl_results import SUPER_BOWL_RESULTS, SuperBowlResult
from tests.fixtures.number_name import NumberName, NUMBER_NAMES


# noinspection PyUnresolvedReferences
class StoreTstABC(ABC):
    """Tests which expect store to have the bands data loaded"""

    @abstractmethod
    def new_super_bowl_results_store(self) -> StoreABC:
        """Create a new store object containing only Superbowl results"""

    @abstractmethod
    def new_number_name_store(self) -> StoreABC:
        """Create a new store object containing only Number Names"""

    def test_read(self):
        store = self.new_super_bowl_results_store()
        found = store.read("iii")
        expected = marshy.load(
            SuperBowlResult,
            {
                "code": "iii",
                "year": 1969,
                "date": "1969-01-12T00:00:00+00:00",
                "winner_code": "new_york_jets",
                "runner_up_code": "baltimore",
                "winner_score": 16,
                "runner_up_score": 7,
            },
        )
        self.assertEqual(expected, found)

    def test_read_not_existing(self):
        store = self.new_super_bowl_results_store()
        found = store.read("not_a_code")
        self.assertIsNone(found)

    def test_read_all(self):
        store = self.new_super_bowl_results_store()
        super_bowl_results = list(store.read_all(("vii", "no-code-i", "ii", "xx")))
        expected = [
            marshy.load(
                SuperBowlResult,
                {
                    "code": "vii",
                    "year": 1973,
                    "date": "1973-01-14T00:00:00+00:00",
                    "winner_code": "miami",
                    "runner_up_code": "washington",
                    "winner_score": 14,
                    "runner_up_score": 7,
                },
            ),
            None,
            marshy.load(
                SuperBowlResult,
                {
                    "code": "ii",
                    "year": 1968,
                    "date": "1968-01-14T00:00:00+00:00",
                    "winner_code": "green_bay",
                    "runner_up_code": "oakland",
                    "winner_score": 33,
                    "runner_up_score": 14,
                },
            ),
            marshy.load(
                SuperBowlResult,
                {
                    "code": "xx",
                    "year": 1986,
                    "date": "1986-01-26T00:00:00+00:00",
                    "winner_code": "chicago",
                    "runner_up_code": "new_england",
                    "winner_score": 46,
                    "runner_up_score": 10,
                },
            ),
        ]
        self.assertEqual(expected, super_bowl_results)

    def test_create(self):
        store = self.new_super_bowl_results_store()
        self.spec_for_create(store)

    def spec_for_create(self, store):
        item = marshy.load(
            SuperBowlResult,
            {
                "code": "c",
                "year": 2067,
                "date": "2067-01-15T00:00:00+00:00",
                "winner_code": "robots",
                "runner_up_code": "humans",
                "winner_score": 1234,
                "runner_up_score": 0,
            },
        )
        self.assertEqual(56, store.count())
        created = store.create(item)
        self.assertEqual(item, created)
        read = store.read("c")
        self.assertEqual(item, read)
        self.assertEqual(57, store.count())

    def test_create_invalid_schema(self):
        store = self.new_super_bowl_results_store()
        item = SuperBowlResult(
            **{
                "code": "c",
                "year": "not_a_year",
                "date": datetime.fromisoformat("2067-01-15T00:00:00+00:00"),
                "winner_code": "robots",
                "runner_up_code": "humans",
                "winner_score": 1234,
                "runner_up_score": 0,
            }
        )
        try:
            store.create(deepcopy(item)) and self.assertTrue(False)
        except PersistyError:
            self.assertEqual(56, store.count())

    def test_create_existing_key(self):
        store = self.new_super_bowl_results_store()
        item = marshy.load(
            SuperBowlResult,
            {
                "code": "i",
                "year": 1971,
                "date": "1967-01-15T00:00:00+00:00",
                "winner_code": "green_bay",
                "runner_up_code": "kansas_city",
                "winner_score": 35,
                "runner_up_score": 10,
            },
        )
        try:
            store.create(deepcopy(item)) and self.assertTrue(False)
        except PersistyError:
            self.assertEqual(56, store.count())

    def test_update(self):
        store = self.new_super_bowl_results_store()
        self.spec_for_update(store)

    def spec_for_update(self, store):
        item = marshy.load(
            SuperBowlResult,
            {
                "code": "li",
                "winner_code": "tom_brady_fan_club",
            },
        )
        updated = store.update(item)
        expected = marshy.load(
            SuperBowlResult,
            {
                "code": "li",
                "year": 2017,
                "date": "2017-02-05T00:00:00+00:00",
                "winner_code": "tom_brady_fan_club",
                "runner_up_code": "atlanta",
                "winner_score": 34,
                "runner_up_score": 28,
            },
        )
        self.assertEqual(expected, updated)
        self.assertEqual(56, store.count())
        item = store.read("li")
        self.assertEqual(expected, item)

    def test_update_missing_key(self):
        store = self.new_super_bowl_results_store()
        self.spec_for_update_missing_key(store)

    def spec_for_update_missing_key(self, store):
        item = SuperBowlResult(
            code="not_a_key",
            year=1971,
        )
        # noinspection PyDataclass
        updates = replace(item)
        self.assertIsNone(store.update(updates))
        self.assertEqual(56, store.count())
        self.assertEqual(item, updates)

    def test_update_invalid_schema(self):
        store = self.new_super_bowl_results_store()
        with self.assertRaises(AttributeError):
            store.update(SuperBowlResult(code="i", date="not_a_date"))
        self.assertEqual(56, store.count())
        read = store.read("i")
        expected = marshy.load(
            SuperBowlResult,
            {
                "code": "i",
                "year": 1967,
                "date": "1967-01-15T00:00:00+00:00",
                "winner_code": "green_bay",
                "runner_up_code": "kansas_city",
                "winner_score": 35,
                "runner_up_score": 10,
            },
        )
        self.assertEqual(56, store.count())
        self.assertEqual(expected, read)

    def test_update_valid_filter(self):
        store = self.new_super_bowl_results_store()
        item = SuperBowlResult(
            code="li",
            winner_code="tom_brady_fan_club",
        )
        self.assertTrue(store.update(item, AttrFilter("year", AttrFilterOp.eq, 2017)))
        item = store.read("li")
        expected = marshy.load(
            SuperBowlResult,
            {
                "code": "li",
                "year": 2017,
                "date": "2017-02-05T00:00:00+00:00",
                "winner_code": "tom_brady_fan_club",
                "runner_up_code": "atlanta",
                "winner_score": 34,
                "runner_up_score": 28,
            },
        )
        self.assertEqual(expected, item)

    def test_update_invalid_filter(self):
        store = self.new_super_bowl_results_store()
        item = SuperBowlResult(code="li", winner_code="tom_brady_fan_club")
        self.assertIsNone(store.update(item, AttrFilter("year", AttrFilterOp.eq, 2018)))
        item = store.read("li")
        expected = marshy.load(
            SuperBowlResult,
            {
                "code": "li",
                "year": 2017,
                "date": "2017-02-05T00:00:00+00:00",
                "winner_code": "new_england",
                "runner_up_code": "atlanta",
                "winner_score": 34,
                "runner_up_score": 28,
            },
        )
        self.assertEqual(expected, item)

    def test_delete(self):
        store = self.new_super_bowl_results_store()
        self.assertTrue(store.delete("lvi"))
        self.assertIsNone(store.read("lvi"))

    def test_delete_missing_key(self):
        store = self.new_super_bowl_results_store()
        self.assertFalse(store.delete("missing_key"))

    def test_count(self):
        store = self.new_super_bowl_results_store()
        self.assertEqual(56, store.count())
        filters = filter_factory(SuperBowlResult)
        self.assertEqual(
            20, store.count(filters.year.gte(1984) & filters.year.lt(2004))
        )
        self.assertEqual(6, store.count(filters.winner_code.contains("new_england")))
        self.assertEqual(0, store.count(AttrFilter("year", AttrFilterOp.lt, 1967)))

    def test_count_invalid_attr_filter(self):
        store = self.new_super_bowl_results_store()
        try:
            store.count(
                AttrFilter("non_attr", AttrFilterOp.gte, 1984)
            ) and self.assertTrue(False)
        except PersistyError:
            pass

    def test_count_exclude_all(self):
        store = self.new_super_bowl_results_store()
        self.assertEqual(0, store.count(EXCLUDE_ALL))

    def test_count_filter_id(self):
        store = self.new_number_name_store()
        filters = filter_factory(NumberName)
        num = store.count(
            filters.id.eq(UUID("00000000-0000-0000-0000-000000000010"))
            & filters.title.eq("Ten")
        )
        self.assertEqual(1, num)

    def test_count_custom_filter(self):
        @dataclass
        class StrLenFilter(SearchFilterABC[T]):
            attr_name: str
            required_length: int

            def lock_attrs(self, attrs: Tuple[Attr, ...]) -> SearchFilterABC[T]:
                attr = next(attr for attr in attrs if attr.name == self.attr_name)
                assert attr.readable
                return self

            def match(self, item: T, attrs: Tuple[Attr, ...]) -> bool:
                # noinspection PyTypeChecker
                return len(getattr(item, self.attr_name)) == self.required_length

        store = self.new_number_name_store()

        num = store.count(StrLenFilter("title", 4))
        self.assertEqual(3, num)  # Four, Five, Nine

    def test_count_non_indexed_filter(self):
        tag_store = self.new_number_name_store()
        filters = filter_factory(NumberName)
        count = tag_store.count(filters.title.eq("Five") & filters.value.eq(5))
        self.assertEqual(1, count)

    def test_search_all(self):
        store = self.new_super_bowl_results_store()
        self.assertEqual(SUPER_BOWL_RESULTS, list(store.search_all()))
        filters = filter_factory(SuperBowlResult)
        self.assertEqual(
            SUPER_BOWL_RESULTS[17:37],
            list(store.search_all(filters.year.gte(1984) & filters.year.lt(2004))),
        )
        self.assertEqual(
            [r for r in SUPER_BOWL_RESULTS if r.winner_code == "new_england"],
            list(
                store.search_all(
                    AttrFilter("winner_code", AttrFilterOp.contains, "new_england")
                )
            ),
        )
        self.assertEqual(
            [], list(store.search_all(AttrFilter("year", AttrFilterOp.lt, 1967)))
        )

    def test_search(self):
        store = self.new_number_name_store()
        filters = filter_factory(NumberName)
        search_filter = filters.value.lt(50)
        search_order = filters.value.asc()
        page1 = store.search(search_filter, search_order, None, 10)
        self.assertEqual(NUMBER_NAMES[:10], page1.results)
        page2 = store.search(search_filter, search_order, page1.next_page_key, 10)
        self.assertEqual(NUMBER_NAMES[10:20], page2.results)

    def test_search_all_id(self):
        store = self.new_number_name_store()
        filters = filter_factory(NumberName)
        id_ = "00000000-0000-0000-0000-000000000001"
        loaded = list(store.search_all(filters.id.eq(UUID(id_))))
        expected = [store.read(id_)]
        self.assertEqual(expected, loaded)

    def test_search_all_id_title(self):
        store = self.new_number_name_store()
        filters = filter_factory(NumberName)
        id_ = "00000000-0000-0000-0000-000000000001"
        loaded = list(
            store.search_all(filters.id.eq(UUID(id_)) & filters.title.contains("One"))
        )
        expected = [store.read(id_)]
        self.assertEqual(expected, loaded)

    def test_edit_all(self):
        store = self.new_number_name_store()
        edits = store.search_all(AttrFilter("value", AttrFilterOp.gt, 3))
        edits = (BatchEdit(delete_key=str(n.id)) for n in edits)
        list(store.edit_all(edits))
        self.assertEqual(3, store.count())
        edits = [
            BatchEdit(
                create_item=NumberName(
                    id=UUID("00000000-0000-0000-0001-000000000000"),
                    title="Minus One",
                    value=-1,
                )
            ),
            BatchEdit(
                update_item=NumberName(
                    id=UUID("00000000-0000-0000-0002-000000000001"),
                    title="Not existing",
                )
            ),
            BatchEdit(
                update_item=NumberName(
                    id=UUID("00000000-0000-0000-0000-000000000001"), title="First"
                )
            ),
            BatchEdit(delete_key="00000000-0000-0000-0002-000000000001"),
        ]
        now = datetime.now().astimezone(timezone.utc) + relativedelta(seconds=-2)
        results = [r.success for r in store.edit_all(edits)]
        self.assertEqual(results, [True, False, True, False])
        results = list(
            store.search_all(search_order=SearchOrder((SearchOrderAttr("value"),)))
        )
        results = sorted(results, key=lambda r: r.value)
        self.assertGreaterEqual(results[0].created_at, now)
        self.assertGreaterEqual(results[0].updated_at, now)
        self.assertLess(results[1].created_at, now)
        self.assertGreaterEqual(results[1].updated_at, now)
        expected_results = [
            NumberName(
                created_at=results[0].created_at,
                id=UUID("00000000-0000-0000-0001-000000000000"),
                title="Minus One",
                updated_at=results[0].updated_at,
                value=-1,
            ),
            NumberName(
                created_at=datetime.fromisoformat("1970-01-01T00:00:00+00:00"),
                id=UUID("00000000-0000-0000-0000-000000000001"),
                title="First",
                updated_at=results[1].updated_at,
                value=1,
            ),
            NumberName(
                created_at=datetime.fromisoformat("1970-01-01T00:00:00+00:00"),
                id=UUID("00000000-0000-0000-0000-000000000002"),
                title="Two",
                updated_at=datetime.fromisoformat("1970-01-01T00:00:00+00:00"),
                value=2,
            ),
            NumberName(
                created_at=datetime.fromisoformat("1970-01-01T00:00:00+00:00"),
                id=UUID("00000000-0000-0000-0000-000000000003"),
                title="Three",
                updated_at=datetime.fromisoformat("1970-01-01T00:00:00+00:00"),
                value=3,
            ),
        ]
        self.assertEqual(expected_results, results)

    def test_edit_batch(self):
        store = self.new_number_name_store()
        edits = [BatchEdit(delete_key=str(n.id)) for n in store.search().results]
        results = store.edit_batch(edits)
        for result in results:
            self.assertTrue(result.success)
        self.assertEqual(89, store.count())

    def test_update_no_key(self):
        store = self.new_number_name_store()
        with self.assertRaises(PersistyError):
            store.update(NumberName())

    def test_update_fail_filter(self):
        store = self.new_number_name_store()
        self.spec_for_update_fail_filter(store)

    def spec_for_update_fail_filter(self, store):
        item = store.update(
            NumberName(
                id=UUID("00000000-0000-0000-0000-000000000001"), title="Not One"
            ),
            filter_factory(NumberName).title.ne("One"),
        )
        self.assertIsNone(item)

    def test_search_custom_filter_full_result_set(self):
        store = self.new_number_name_store()
        search_filter = ValueLessThanFilter(21)
        page_1 = store.search(search_filter)
        self.assertEqual(list(range(1, 11)), list(i.value for i in page_1.results))
        page_2 = store.search(
            search_filter=search_filter, page_key=page_1.next_page_key
        )
        self.assertEqual(list(range(11, 21)), list(i.value for i in page_2.results))
        page_3 = store.search(
            search_filter=search_filter, page_key=page_2.next_page_key
        )
        self.assertEqual(ResultSet([]), page_3)

    def test_search_custom_filter_unfilled_result_set(self):
        store = self.new_number_name_store()
        limit = 3
        for less_than in range(1, 31):
            kwargs = dict(
                search_filter=ValueLessThanFilter(less_than),
                search_order=SearchOrder((SearchOrderAttr("value"),)),
                limit=limit,
            )
            index = 1
            while True:
                page = store.search(**kwargs)
                expected_values = [
                    v for v in range(index, min(less_than, index + limit))
                ]
                values = [r.value for r in page.results]
                self.assertEqual(expected_values, values)
                if page.next_page_key:
                    kwargs["page_key"] = page.next_page_key
                    index += limit
                else:
                    break

    def test_edit_batch_errors(self):
        store = self.new_number_name_store()
        edits = [
            BatchEdit(
                create_item=NumberName(
                    id=NUMBER_NAMES[1].id, value=-1, title="New Item"
                )
            ),
            BatchEdit(
                update_item=NumberName(id=uuid4(), value=-2, title="Updated Item")
            ),
            BatchEdit(delete_key=str(uuid4())),
            BatchEdit(),
        ]
        results = store.edit_batch(edits)
        self.assertFalse(next((True for r in results if r.success), False))

    def test_wrong_type_str(self):
        store = self.new_number_name_store()
        id = uuid4()
        item = NumberName(
            id=str(id),  # The type is wrong here, but we should handle it
            title="A Gazillion",
            value=-1,
        )
        store.create(item)
        # noinspection PyTypeChecker
        loaded = store.read(id)  # Again - type is wrong - this should be a str
        self.assertIsNotNone(loaded.created_at)
        self.assertIsNotNone(loaded.updated_at)
        item.id = id
        item.created_at = loaded.created_at
        item.updated_at = loaded.updated_at
        self.assertEqual(item, loaded)

        filters = filter_factory(NumberName)
        search_filter = filters.id.eq(str(id))
        results = list(store.search_all(search_filter))
        self.assertEqual([item], results)

        # noinspection PyTypeChecker
        self.assertTrue(
            store.delete(id)
        )  # Again - type is wrong - this should be a str
        # noinspection PyTypeChecker
        self.assertFalse(store.read(id))  # Again - type is wrong - this should be a str


@dataclass
class ValueLessThanFilter(SearchFilterABC[T]):
    """
    Custom filter for testing - in reality you would use a AttrFilter
    for this as store implementations would more easily be able to turn
    it into a native condition
    """

    value: int

    def lock_attrs(self, attrs: Tuple[Attr, ...]) -> SearchFilterABC[T]:
        return self

    def match(self, item: T, attrs: Tuple[Attr, ...]) -> bool:
        return item.value < self.value



from decimal import Decimal
from typing import List
from unittest import TestCase

from marshy import dump
from marshy.types import ExternalItemType

from persisty.attr.attr import Attr
from persisty.attr.attr_type import AttrType
from persisty.attr.generator.default_value_generator import DefaultValueGenerator
from persisty.impl.dynamodb.partition_sort_index import PartitionSortIndex
from persisty.impl.dynamodb.dynamodb_store_factory import DynamodbStoreFactory
from persisty.key_config.attr_key_config import AttrKeyConfig
from persisty.key_config.composite_key_config import CompositeKeyConfig
from persisty.result_set import ResultSet

from persisty.search_filter.exclude_all import EXCLUDE_ALL
from persisty.search_filter.filter_factory import filter_factory

from persisty.store.store_abc import StoreABC
from persisty.store.wrapper_store_abc import WrapperStoreABC
from persisty.store_meta import get_meta
from persisty.stored import stored
from tests.fixtures.number_name import NumberName, NUMBER_NAMES_DICTS
from tests.fixtures.storage_tst_abc import StoreTstABC
from tests.fixtures.super_bowl_results import SuperBowlResult, SUPER_BOWL_RESULT_DICTS
from tests.utils import mock_dynamodb_with_super


@mock_dynamodb_with_super
class TestDynamodbStore(TestCase, StoreTstABC):
    def new_super_bowl_results_store(self) -> StoreABC:
        store_factory = DynamodbStoreFactory(meta=get_meta(SuperBowlResult))
        self.seed_table(store_factory, SUPER_BOWL_RESULT_DICTS)
        store = store_factory.create()
        return store

    def new_number_name_store(self) -> StoreABC:
        store_factory = DynamodbStoreFactory(meta=get_meta(NumberName))
        self.seed_table(store_factory, NUMBER_NAMES_DICTS)
        store = store_factory.create()
        return store

    @staticmethod
    def seed_table(store_factory: DynamodbStoreFactory, items: List[ExternalItemType]):
        store_factory.derive_from_meta()
        store_factory.create_table_in_aws()
        dynamodb = store_factory.get_session().resource("dynamodb")
        table = dynamodb.Table(store_factory.table_name)
        with table.batch_writer() as batch:
            for result in items:
                batch.put_item(result)

    def test_update_with_sk(self):
        tag_store = self.new_tag_store()
        tag = Tag(pk=3, sk=5, title="Updated")
        key = tag_store.get_meta().key_config.to_key_str(tag)
        item = tag_store.update(tag)
        tag.codes = ["5"]
        self.assertEqual(round(1 / item.weight), 305)
        tag.weight = item.weight
        self.assertEqual(tag, item)
        loaded = tag_store.read(key)
        self.assertEqual(loaded, tag)

    def test_dynamodb_update_fail_filter(self):
        # noinspection PyUnresolvedReferences
        store = self.new_number_name_store().store
        self.spec_for_update_fail_filter(store)

    def test_read_with_sk_missing(self):
        tag_store = self.new_tag_store()
        key = Tag(pk=3, sk=5000)
        key = tag_store.get_meta().key_config.to_key_str(key)
        item = tag_store.read(key)
        self.assertIsNone(item)

    def test_search_exclude_all(self):
        # noinspection PyTypeChecker
        tag_store: WrapperStoreABC = self.new_tag_store()
        result_set = tag_store.get_store().search(EXCLUDE_ALL)
        self.assertEqual(ResultSet([]), result_set)

    def test_create_with_float(self):
        tag_store = self.new_tag_store()
        tag = Tag(
            pk=20,
            sk=20,
            title="Two thousand and twenty",
            codes=["foo", "bar"],
            weight=0.2,
        )
        tag_store.create(tag)
        key = tag_store.get_meta().key_config.to_key_str(tag)
        results = list(tag_store.search_all())
        self.assertEqual(1001, len(results))
        item = tag_store.read(key)
        self.assertEqual(item, tag)

    def test_create_with_float_int(self):
        tag_store = self.new_tag_store()
        tag = Tag(
            pk=20,
            sk=20,
            title="Two thousand and twenty",
            codes=["foo", "bar"],
            weight=1,
        )
        tag_store.create(tag)
        key = tag_store.get_meta().key_config.to_key_str(Tag(pk=20, sk=20))
        item = tag_store.read(key)
        self.assertEqual(item, tag)

    def test_count_multi_page(self):
        tag_store = self.new_tag_store()
        filters = filter_factory(Tag)
        self.assertEqual(1000, tag_store.count(filters.title.ne("foobar")))
        self.assertEqual(10, tag_store.count(filters.sk.eq(10)))

    def test_convert_to_decimals(self):
        item = {"some_int": 10, "some_float": 0.5}
        # noinspection PyUnresolvedReferences
        store = self.new_tag_store()
        # noinspection PyUnresolvedReferences
        store = store.store
        converted = store._convert_to_decimals(item)
        expected = {"some_int": Decimal("10"), "some_float": Decimal(0.5)}
        self.assertEqual(expected, converted)

    def test_dammit(self):
        self.test_create_with_float()

    def new_tag_store(self) -> StoreABC:
        store_meta = get_meta(Tag)
        store_factory = DynamodbStoreFactory(
            meta=store_meta,
            index=PartitionSortIndex("pk", "sk"),
            global_secondary_indexes=dict(gix__sk__pk=PartitionSortIndex("sk", "pk")),
        )
        tags = list(
            dump(
                Tag(
                    int(i / 100),
                    i % 100,
                    str(i),
                    [str(c) for c in (2, 3, 5) if not i % c],
                    Decimal("%.9f" % (1 / i)),
                )
            )
            for i in range(1, 1001)
        )
        self.seed_table(store_factory, tags)
        store = store_factory.create()
        return store


@stored(
    batch_size=10,
    key_config=CompositeKeyConfig(
        (AttrKeyConfig("pk", AttrType.INT), AttrKeyConfig("sk", AttrType.INT))
    ),
)
class Tag:
    pk: int
    sk: int
    title: str
    codes: List[str] = Attr(create_generator=DefaultValueGenerator([]))
    weight: float = 0.5



from marshy.types import ExternalItemType
from servey.servey_aws.serverless.yml_config.yml_config_abc import YmlConfigABC, ensure_ref_in_file, create_yml_file

from persisty.finder.store_finder_abc import find_stores
from persisty.impl.default_store import DefaultStore
from persisty.impl.dynamodb.dynamodb_store_factory import DynamodbStoreFactory
from persisty.impl.dynamodb.dynamodb_table_store import DynamodbTableStore


class DynamodbYmlConfig(YmlConfigABC):
    """
    Set up some aspect of the serverless environment yml files. (For example, functions, resources, etc...)
    """

    dynamodb_resource_yml_file: str = "serverless_servey/dynamodb_resource.yml"
    dynamodb_role_statement_yml_file: str = "serverless_servey/dynamodb_role_statement.yml"

    def configure(self, main_serverless_yml_file: str):

        ensure_ref_in_file(
            main_serverless_yml_file,
            ["resources"],
            self.dynamodb_resource_yml_file,
        )
        ensure_ref_in_file(
            main_serverless_yml_file,
            ["provider", "iamRoleStatements"],
            self.dynamodb_role_statement_yml_file,
            "iamRoleStatements.0",
        )
        dynamodb_resource_yml = self.build_dynamodb_resource_yml()
        create_yml_file(self.dynamodb_resource_yml_file, dynamodb_resource_yml)
        dynamodb_role_statement_yml = self.build_dynamodb_role_statement_yml()
        create_yml_file(self.dynamodb_role_statement_yml_file, dynamodb_role_statement_yml)

    @staticmethod
    def get_dynamodb_store_factories():
        stores = find_stores()
        for store in stores:
            if isinstance(store, DynamodbTableStore) or isinstance(store, DefaultStore):
                factory = DynamodbStoreFactory(store.meta)
                factory.derive_from_meta()
                yield factory

    def build_dynamodb_resource_yml(self) -> ExternalItemType:
        resources = {}
        for factory in self.get_dynamodb_store_factories():
            resources[factory.table_name.title().replace('_', '')] = {
                "Type": "AWS::DynamoDB::Table",
                "Properties": {
                    "TableName": factory.table_name,
                    "AttributeDefinitions": factory.get_attribute_definitions(),
                    "KeySchema": factory.index.to_schema(),
                    "GlobalSecondaryIndexes": factory.get_global_secondary_indexes(),
                    "BillingMode": "PAY_PER_REQUEST"
                }
            }
        return {
            "Resources": resources
        }

    def build_dynamodb_role_statement_yml(self) -> ExternalItemType:
        resources = []
        for factory in self.get_dynamodb_store_factories():
            resource_name = factory.table_name.title().replace('_', '')
            resources.append({"Fn::GetAtt": [resource_name, "Arn"]})
            for index_name, index in factory.global_secondary_indexes.items():
                resources.append({
                    "Fn::Join": ["/", [{"Fn::GetAtt": [resource_name, "Arn"]}, "index", index_name]]
                })
        iam_role_statement = {
            "Effect": "Allow",
            "Action": [
                "dynamodb:DescribeTable",
                "dynamodb:Query",
                "dynamodb:Scan",
                "dynamodb:BatchGetItem",
                "dynamodb:GetItem",
                "dynamodb:PutItem",
                "dynamodb:UpdateItem",
                "dynamodb:DeleteItem",
            ],
            "Resource": resources
        }
        return {
            "iamRoleStatements": [iam_role_statement]
        }

    @staticmethod
    def _iam_role_statement(resource):
        result = {
            "Effect": "Allow",
            "Action": [
                "dynamodb:DescribeTable",
                "dynamodb:Query",
                "dynamodb:Scan",
                "dynamodb:BatchGetItem",
                "dynamodb:GetItem",
                "dynamodb:PutItem",
                "dynamodb:UpdateItem",
                "dynamodb:DeleteItem",
            ],
            "Resource": resource
        }
        return result



from unittest import TestCase

from persisty.util import UNDEFINED
from persisty.util.undefined import Undefined


class TestUndefined(TestCase):
    def test_bool(self):
        self.assertFalse(bool(UNDEFINED))

    def test_hash(self):
        self.assertEqual(hash(UNDEFINED), hash(Undefined()))

    def test_singleton(self):
        self.assertIs(UNDEFINED, Undefined())



from dataclasses import dataclass
from unittest import TestCase
from unittest.mock import patch

from persisty.factory.default_store_factory import DefaultStoreFactory
from persisty.finder.module_store_finder import ModuleStoreFinder
from persisty.finder.store_finder_abc import find_stores, find_store_factories


class TestModuleStoreFinder(TestCase):
    def test_find_stores(self):
        finder = ModuleStoreFinder("tests.finder")
        stores = list(finder.find_stores())
        store_names = {s.get_meta().name for s in stores}
        expected_store_names = {"message"}
        self.assertEqual(store_names, expected_store_names)

    def test_find_store_factories(self):
        finder = ModuleStoreFinder("tests.finder")
        factories = list(finder.find_store_factories())
        store_names = {s.get_meta().name for s in factories}
        expected_store_names = {"message"}
        self.assertEqual(store_names, expected_store_names)
        factory = next(iter(factories))
        self.assertTrue(isinstance(factory, DefaultStoreFactory))

    def test_globals(self):
        @dataclass
        class MyModuleStoreFinder(ModuleStoreFinder):
            root_module_name: str = "tests.finder"

        with patch(
            "persisty.finder.store_finder_abc.get_impls",
            return_value=[MyModuleStoreFinder],
        ):
            expected_store_names = {"message"}
            stores = list(find_stores())
            factories = list(find_store_factories())
            self.assertEqual({s.get_meta().name for s in stores}, expected_store_names)
            self.assertEqual(
                {s.get_meta().name for s in factories}, expected_store_names
            )



from unittest import TestCase

import marshy

from persisty.attr.attr_filter_op import (
    _lte,
    _endswith,
    _oneof,
    _exists,
    _not_exists,
    AttrFilterOp,
)
from persisty.util import UNDEFINED


class TestAttrFilterOp(TestCase):
    def test_lte(self):
        self.assertTrue(_lte(5, 6))
        self.assertTrue(_lte(6, 6))
        self.assertFalse(_lte(7, 6))

    def test_endswith(self):
        self.assertTrue(_endswith("foobar", "bar"))
        self.assertFalse(_endswith("foobarg", "bar"))

    def test_oneof(self):
        self.assertTrue(_oneof("a", ("a", "b", "c")))
        self.assertFalse(_oneof("d", ("a", "b", "c")))

    def test_exists(self):
        self.assertFalse(_exists(None, 1))
        self.assertFalse(_exists(UNDEFINED, 1))
        self.assertTrue(_exists("foobar", 1))

    def test_not_exists(self):
        self.assertTrue(_not_exists(None, 1))
        self.assertTrue(_not_exists(UNDEFINED, 1))
        self.assertFalse(_not_exists("foobar", 1))

    def test_attr_filter_op_marshaller(self):
        self.assertEqual("eq", marshy.dump(AttrFilterOp.eq))
        self.assertEqual("lte", marshy.dump(AttrFilterOp.lte))
        self.assertEqual(AttrFilterOp.eq, marshy.load(AttrFilterOp, "eq"))
        self.assertEqual(AttrFilterOp.lte, marshy.load(AttrFilterOp, "lte"))



from typing import List
from unittest import TestCase

from dataclasses import dataclass, field

from servey.util.singleton_abc import SingletonABC

from persisty.util.logify import logify


# noinspection PyMethodMayBeStatic
class Greeter(SingletonABC):
    def say_hello(self, name: str) -> str:
        return f"Hello {name}"

    def say_hello_kwargs(self, *, name: str) -> str:
        return f"Hello {name}"

    def say_hello_args(self, name: str, **_) -> str:
        return f"Hello {name}"

    def raise_an_error(self):
        raise TypeError("i_dont_like_you")


@dataclass
class MockLogger:
    infos: List = field(default_factory=list)

    def info(self, msg):
        self.infos.append(msg)


class TestLogify(TestCase):
    def test_logify(self):
        logger = MockLogger()
        # noinspection PyTypeChecker
        greeter = logify(Greeter(), logger=logger)
        greeter.say_hello("Developer")
        expected = MockLogger(
            [
                {
                    "name": "say_hello",
                    "time": 0,
                    "args": ("Developer",),
                }
            ]
        )
        self.assertEqual(expected, logger)

    def test_logify_args(self):
        logger = MockLogger()
        # noinspection PyTypeChecker
        greeter = logify(Greeter(), logger=logger)
        greeter.say_hello_args("Developer")
        expected = MockLogger(
            [
                {
                    "name": "say_hello_args",
                    "time": 0,
                    "args": ("Developer",),
                }
            ]
        )
        self.assertEqual(expected, logger)

    def test_logify_kwargs(self):
        logger = MockLogger()
        # noinspection PyTypeChecker
        greeter = logify(Greeter(), logger=logger)
        greeter.say_hello_kwargs(name="Developer")
        expected = MockLogger(
            [
                {
                    "name": "say_hello_kwargs",
                    "time": 0,
                    "kwargs": {"name": "Developer"},
                }
            ]
        )
        self.assertEqual(expected, logger)

    def test_logify_error(self):
        logger = MockLogger()
        # noinspection PyTypeChecker
        greeter = logify(Greeter(), logger=logger)
        try:
            greeter.raise_an_error() and self.assertTrue(False)
        except TypeError as e:
            self.assertEqual(str(e), "i_dont_like_you")  # He doesn't like you either!
        expected = MockLogger(
            [
                {
                    "name": "raise_an_error",
                    "time": 0,
                    "error": "i_dont_like_you",
                }
            ]
        )
        self.assertEqual(expected, logger)

    def test_logify_skip(self):
        logger = MockLogger()
        # noinspection PyTypeChecker
        greeter = logify(
            Greeter(), log_methods=frozenset(("say_hello_args",)), logger=logger
        )
        greeter.say_hello("Developer")
        expected = MockLogger()
        self.assertEqual(expected, logger)

    def test_logify_default(self):
        logger = MockLogger()
        # noinspection PyTypeChecker
        greeter = logify(Greeter())
        greeter.say_hello("Developer")



from datetime import datetime

from schemey.schema import int_schema

from persisty.attr.attr import Attr
from persisty.attr.attr_type import AttrType
from persisty.key_config.attr_key_config import AttrKeyConfig
from persisty.stored import stored

SUPER_BOWL_RESULT_DICTS = [
    {
        "code": "i",
        "year": 1967,
        "date": "1967-01-15T00:00:00+00:00",
        "winner_code": "green_bay",
        "runner_up_code": "kansas_city",
        "winner_score": 35,
        "runner_up_score": 10,
    },
    {
        "code": "ii",
        "year": 1968,
        "date": "1968-01-14T00:00:00+00:00",
        "winner_code": "green_bay",
        "runner_up_code": "oakland",
        "winner_score": 33,
        "runner_up_score": 14,
    },
    {
        "code": "iii",
        "year": 1969,
        "date": "1969-01-12T00:00:00+00:00",
        "winner_code": "new_york_jets",
        "runner_up_code": "baltimore",
        "winner_score": 16,
        "runner_up_score": 7,
    },
    {
        "code": "iv",
        "year": 1970,
        "date": "1970-01-11T00:00:00+00:00",
        "winner_code": "kansas_city",
        "runner_up_code": "minnesota",
        "winner_score": 23,
        "runner_up_score": 7,
    },
    {
        "code": "v",
        "year": 1971,
        "date": "1971-01-17T00:00:00+00:00",
        "winner_code": "baltimore",
        "runner_up_code": "dallas",
        "winner_score": 16,
        "runner_up_score": 13,
    },
    {
        "code": "vi",
        "year": 1972,
        "date": "1972-01-16T00:00:00+00:00",
        "winner_code": "dallas",
        "runner_up_code": "miami",
        "winner_score": 24,
        "runner_up_score": 3,
    },
    {
        "code": "vii",
        "year": 1973,
        "date": "1973-01-14T00:00:00+00:00",
        "winner_code": "miami",
        "runner_up_code": "washington",
        "winner_score": 14,
        "runner_up_score": 7,
    },
    {
        "code": "viii",
        "year": 1974,
        "date": "1974-01-13T00:00:00+00:00",
        "winner_code": "miami",
        "runner_up_code": "minnesota",
        "winner_score": 24,
        "runner_up_score": 7,
    },
    {
        "code": "ix",
        "year": 1975,
        "date": "1975-01-12T00:00:00+00:00",
        "winner_code": "pittsburgh",
        "runner_up_code": "minnesota",
        "winner_score": 16,
        "runner_up_score": 6,
    },
    {
        "code": "x",
        "year": 1976,
        "date": "1976-01-18T00:00:00+00:00",
        "winner_code": "pittsburgh",
        "runner_up_code": "dallas",
        "winner_score": 21,
        "runner_up_score": 17,
    },
    {
        "code": "xi",
        "year": 1977,
        "date": "1977-01-09T00:00:00+00:00",
        "winner_code": "oakland",
        "runner_up_code": "minnesota",
        "winner_score": 32,
        "runner_up_score": 14,
    },
    {
        "code": "xii",
        "year": 1978,
        "date": "1978-01-15T00:00:00+00:00",
        "winner_code": "dallas",
        "runner_up_code": "denver",
        "winner_score": 27,
        "runner_up_score": 10,
    },
    {
        "code": "xiii",
        "year": 1979,
        "date": "1979-01-21T00:00:00+00:00",
        "winner_code": "pittsburgh",
        "runner_up_code": "dallas",
        "winner_score": 35,
        "runner_up_score": 31,
    },
    {
        "code": "xiv",
        "year": 1980,
        "date": "1980-01-20T00:00:00+00:00",
        "winner_code": "pittsburgh",
        "runner_up_code": "los_angeles_rams",
        "winner_score": 31,
        "runner_up_score": 19,
    },
    {
        "code": "xv",
        "year": 1981,
        "date": "1981-01-25T00:00:00+00:00",
        "winner_code": "oakland",
        "runner_up_code": "philadelphia",
        "winner_score": 27,
        "runner_up_score": 10,
    },
    {
        "code": "xvi",
        "year": 1982,
        "date": "1982-01-24T00:00:00+00:00",
        "winner_code": "san_francisco",
        "runner_up_code": "cincinnati",
        "winner_score": 26,
        "runner_up_score": 21,
    },
    {
        "code": "xvii",
        "year": 1983,
        "date": "1983-01-30T00:00:00+00:00",
        "winner_code": "washington",
        "runner_up_code": "miami",
        "winner_score": 27,
        "runner_up_score": 17,
    },
    {
        "code": "xviii",
        "year": 1984,
        "date": "1984-01-22T00:00:00+00:00",
        "winner_code": "los_angeles_raiders",
        "runner_up_code": "washington",
        "winner_score": 38,
        "runner_up_score": 9,
    },
    {
        "code": "xix",
        "year": 1985,
        "date": "1985-01-20T00:00:00+00:00",
        "winner_code": "san_francisco",
        "runner_up_code": "miami",
        "winner_score": 38,
        "runner_up_score": 16,
    },
    {
        "code": "xx",
        "year": 1986,
        "date": "1986-01-26T00:00:00+00:00",
        "winner_code": "chicago",
        "runner_up_code": "new_england",
        "winner_score": 46,
        "runner_up_score": 10,
    },
    {
        "code": "xxi",
        "year": 1987,
        "date": "1987-01-25T00:00:00+00:00",
        "winner_code": "new_york_giants",
        "runner_up_code": "denver",
        "winner_score": 39,
        "runner_up_score": 20,
    },
    {
        "code": "xxii",
        "year": 1988,
        "date": "1988-01-31T00:00:00+00:00",
        "winner_code": "washington",
        "runner_up_code": "denver",
        "winner_score": 42,
        "runner_up_score": 10,
    },
    {
        "code": "xxiii",
        "year": 1989,
        "date": "1989-01-22T00:00:00+00:00",
        "winner_code": "san_francisco",
        "runner_up_code": "cincinnati",
        "winner_score": 20,
        "runner_up_score": 16,
    },
    {
        "code": "xxiv",
        "year": 1990,
        "date": "1990-01-28T00:00:00+00:00",
        "winner_code": "san_francisco",
        "runner_up_code": "denver",
        "winner_score": 55,
        "runner_up_score": 10,
    },
    {
        "code": "xxv",
        "year": 1991,
        "date": "1991-01-27T00:00:00+00:00",
        "winner_code": "new_york_giants",
        "runner_up_code": "buffalo",
        "winner_score": 20,
        "runner_up_score": 19,
    },
    {
        "code": "xxvi",
        "year": 1992,
        "date": "1992-01-26T00:00:00+00:00",
        "winner_code": "washington",
        "runner_up_code": "buffalo",
        "winner_score": 37,
        "runner_up_score": 24,
    },
    {
        "code": "xxvii",
        "year": 1993,
        "date": "1993-01-31T00:00:00+00:00",
        "winner_code": "dallas",
        "runner_up_code": "buffalo",
        "winner_score": 52,
        "runner_up_score": 17,
    },
    {
        "code": "xxviii",
        "year": 1994,
        "date": "1994-01-30T00:00:00+00:00",
        "winner_code": "dallas",
        "runner_up_code": "buffalo",
        "winner_score": 30,
        "runner_up_score": 13,
    },
    {
        "code": "xxix",
        "year": 1995,
        "date": "1995-01-29T00:00:00+00:00",
        "winner_code": "san_francisco",
        "runner_up_code": "san_diego",
        "winner_score": 49,
        "runner_up_score": 26,
    },
    {
        "code": "xxx",
        "year": 1996,
        "date": "1996-01-28T00:00:00+00:00",
        "winner_code": "dallas",
        "runner_up_code": "pittsburgh",
        "winner_score": 27,
        "runner_up_score": 17,
    },
    {
        "code": "xxxi",
        "year": 1997,
        "date": "1997-01-26T00:00:00+00:00",
        "winner_code": "green_bay",
        "runner_up_code": "new_england",
        "winner_score": 35,
        "runner_up_score": 21,
    },
    {
        "code": "xxxii",
        "year": 1998,
        "date": "1998-01-25T00:00:00+00:00",
        "winner_code": "denver",
        "runner_up_code": "green_bay",
        "winner_score": 31,
        "runner_up_score": 24,
    },
    {
        "code": "xxxiii",
        "year": 1999,
        "date": "1999-01-31T00:00:00+00:00",
        "winner_code": "denver",
        "runner_up_code": "atlanta",
        "winner_score": 34,
        "runner_up_score": 19,
    },
    {
        "code": "xxxiv",
        "year": 2000,
        "date": "2000-01-30T00:00:00+00:00",
        "winner_code": "st._louis",
        "runner_up_code": "tennessee",
        "winner_score": 23,
        "runner_up_score": 16,
    },
    {
        "code": "xxxv",
        "year": 2001,
        "date": "2001-01-28T00:00:00+00:00",
        "winner_code": "baltimore",
        "runner_up_code": "new_york_giants",
        "winner_score": 34,
        "runner_up_score": 7,
    },
    {
        "code": "xxxvi",
        "year": 2002,
        "date": "2002-02-03T00:00:00+00:00",
        "winner_code": "new_england",
        "runner_up_code": "st._louis",
        "winner_score": 20,
        "runner_up_score": 17,
    },
    {
        "code": "xxxvii",
        "year": 2003,
        "date": "2003-01-26T00:00:00+00:00",
        "winner_code": "tampa_bay",
        "runner_up_code": "oakland",
        "winner_score": 48,
        "runner_up_score": 21,
    },
    {
        "code": "xxxviii",
        "year": 2004,
        "date": "2004-02-01T00:00:00+00:00",
        "winner_code": "new_england",
        "runner_up_code": "carolina",
        "winner_score": 32,
        "runner_up_score": 29,
    },
    {
        "code": "xxxix",
        "year": 2005,
        "date": "2005-02-06T00:00:00+00:00",
        "winner_code": "new_england",
        "runner_up_code": "philadelphia",
        "winner_score": 24,
        "runner_up_score": 21,
    },
    {
        "code": "xl",
        "year": 2006,
        "date": "2006-02-05T00:00:00+00:00",
        "winner_code": "pittsburgh",
        "runner_up_code": "seattle",
        "winner_score": 21,
        "runner_up_score": 10,
    },
    {
        "code": "xli",
        "year": 2007,
        "date": "2007-02-04T00:00:00+00:00",
        "winner_code": "indianapolis",
        "runner_up_code": "chicago",
        "winner_score": 29,
        "runner_up_score": 17,
    },
    {
        "code": "xlii",
        "year": 2008,
        "date": "2008-02-03T00:00:00+00:00",
        "winner_code": "new_york_giants",
        "runner_up_code": "new_england",
        "winner_score": 17,
        "runner_up_score": 14,
    },
    {
        "code": "xliii",
        "year": 2009,
        "date": "2009-02-01T00:00:00+00:00",
        "winner_code": "pittsburgh",
        "runner_up_code": "arizona",
        "winner_score": 27,
        "runner_up_score": 23,
    },
    {
        "code": "xliv",
        "year": 2010,
        "date": "2010-02-07T00:00:00+00:00",
        "winner_code": "new_orleans",
        "runner_up_code": "indianapolis",
        "winner_score": 31,
        "runner_up_score": 17,
    },
    {
        "code": "xlv",
        "year": 2011,
        "date": "2011-02-06T00:00:00+00:00",
        "winner_code": "green_bay",
        "runner_up_code": "pittsburgh",
        "winner_score": 31,
        "runner_up_score": 25,
    },
    {
        "code": "xlvi",
        "year": 2012,
        "date": "2012-02-05T00:00:00+00:00",
        "winner_code": "new_york_giants",
        "runner_up_code": "new_england",
        "winner_score": 21,
        "runner_up_score": 17,
    },
    {
        "code": "xlvii",
        "year": 2013,
        "date": "2013-02-03T00:00:00+00:00",
        "winner_code": "baltimore",
        "runner_up_code": "san_francisco",
        "winner_score": 34,
        "runner_up_score": 31,
    },
    {
        "code": "xlviii",
        "year": 2014,
        "date": "2014-02-02T00:00:00+00:00",
        "winner_code": "seattle",
        "runner_up_code": "denver",
        "winner_score": 43,
        "runner_up_score": 8,
    },
    {
        "code": "xlix",
        "year": 2015,
        "date": "2015-02-01T00:00:00+00:00",
        "winner_code": "new_england",
        "runner_up_code": "seattle",
        "winner_score": 28,
        "runner_up_score": 24,
    },
    {
        "code": "50",
        "year": 2016,
        "date": "2016-02-07T00:00:00+00:00",
        "winner_code": "denver",
        "runner_up_code": "carolina",
        "winner_score": 24,
        "runner_up_score": 10,
    },
    {
        "code": "li",
        "year": 2017,
        "date": "2017-02-05T00:00:00+00:00",
        "winner_code": "new_england",
        "runner_up_code": "atlanta",
        "winner_score": 34,
        "runner_up_score": 28,
    },
    {
        "code": "lii",
        "year": 2018,
        "date": "2018-02-04T00:00:00+00:00",
        "winner_code": "philadelphia",
        "runner_up_code": "new_england",
        "winner_score": 41,
        "runner_up_score": 33,
    },
    {
        "code": "liii",
        "year": 2019,
        "date": "2019-02-03T00:00:00+00:00",
        "winner_code": "new_england",
        "runner_up_code": "los_angeles_rams",
        "winner_score": 13,
        "runner_up_score": 3,
    },
    {
        "code": "liv",
        "year": 2020,
        "date": "2020-02-02T00:00:00+00:00",
        "winner_code": "kansas_city",
        "runner_up_code": "san_francisco",
        "winner_score": 31,
        "runner_up_score": 20,
    },
    {
        "code": "lv",
        "year": 2021,
        "date": "2021-02-07T00:00:00+00:00",
        "winner_code": "tampa_bay",
        "runner_up_code": "kansas_city",
        "winner_score": 31,
        "runner_up_score": 9,
    },
    {
        "code": "lvi",
        "year": 2022,
        "date": "2022-02-13T00:00:00+00:00",
        "winner_code": "los_angeles_rams",
        "runner_up_code": "cincinnati",
        "winner_score": 23,
        "runner_up_score": 20,
    },
]


@stored(key_config=AttrKeyConfig("code", AttrType.STR))
class SuperBowlResult:
    code: str = Attr()
    year: int = Attr(schema=int_schema(minimum=1967), sortable=True)
    date: datetime
    winner_code: str
    runner_up_code: str
    winner_score: int
    runner_up_score: int


SUPER_BOWL_RESULTS = [
    SuperBowlResult(**{**r, "date": datetime.fromisoformat(r["date"])})
    for r in SUPER_BOWL_RESULT_DICTS
]



from typing import Iterator
from unittest import TestCase

from marshy.types import ExternalItemType

from persisty.impl.sqlalchemy.sqlalchemy_context_factory import SqlalchemyContextFactory
from persisty.impl.sqlalchemy.sqlalchemy_table_store_factory import (
    SqlalchemyTableStoreFactory,
)
from persisty.store.store_abc import StoreABC
from persisty.store_meta import StoreMeta, get_meta
from tests.fixtures.number_name import NumberName, NUMBER_NAMES
from tests.fixtures.storage_tst_abc import StoreTstABC
from tests.fixtures.super_bowl_results import (
    SuperBowlResult,
    SUPER_BOWL_RESULTS,
)


class TestSqlalchemyTableStore(TestCase, StoreTstABC):
    def setUp(self) -> None:
        self.context = SqlalchemyContextFactory().create()

    def tearDown(self) -> None:
        pass

    def new_super_bowl_results_store(self) -> StoreABC:
        store_meta = get_meta(SuperBowlResult)
        factory = SqlalchemyTableStoreFactory(store_meta, self.context)
        store = factory.create()
        number_names = ({**r.__dict__, "date": r.date} for r in SUPER_BOWL_RESULTS)
        self.seed_table(store_meta, number_names)
        return store

    def new_number_name_store(self) -> StoreABC:
        store_meta = get_meta(NumberName)
        factory = SqlalchemyTableStoreFactory(store_meta, self.context)
        store = factory.create()
        number_names = (
            {
                **r.__dict__,
                "id": str(r.id),
                "created_at": r.created_at,
                "updated_at": r.updated_at,
            }
            for r in NUMBER_NAMES
        )
        self.seed_table(store_meta, number_names)
        return store

    def seed_table(self, store_meta: StoreMeta, items: Iterator[ExternalItemType]):
        table = self.context.get_table(store_meta)
        with self.context.engine.begin() as conn:
            stmt = table.insert()
            for item in items:
                conn.execute(stmt, item)
            conn.commit()

    """
    def test_search_all_sorted(self):
        store = self.new_super_bowl_results_store()
        filters = filter_factory(SuperBowlResult)
        self.assertEqual(
            list(reversed(SUPER_BOWL_RESULT_DICTS)),
            list(store.search_all(INCLUDE_ALL, filters.year.desc())),
        )
        self.assertEqual(
            list(reversed(SUPER_BOWL_RESULT_DICTS[17:37])),
            list(
                store.search_all(
                    filters.year.gte(1984) & filters.year.lt(2004), filters.year.desc()
                )
            ),
        )

    def test_mem_store_no_dict(self):
        store = mem_store(get_meta(NumberName))
        created = store.create(dict(value=1, title="One"))
        loaded = store.read(created["id"])
        self.assertEqual(loaded, created)

    def test_mem_store_secured(self):
        store_meta = StoreMeta(
            name="read_only_number_name",
            attrs=get_meta(NumberName).attrs,
            access_control=READ_ONLY,
        )
        store = mem_store(store_meta)
        store.read("1")
        error = None
        try:
            store.create(dict(value=-1, title="Minus One"))
        except PersistyError as e:
            error = e
        self.assertIsNotNone(error)

    def test_mem_store_delete_missing_key(self):
        store = MemStore(get_meta(NumberName))
        self.assertFalse(store.delete("missing_key"))

    def test_mem_store_update_missing_key(self):
        # noinspection PyUnresolvedReferences
        store = self.new_super_bowl_results_store().store
        self.spec_for_update_missing_key(store)

    def test_mem_store_update_fail_filter(self):
        # noinspection PyUnresolvedReferences
        store = self.new_number_name_store().store
        self.spec_for_update_fail_filter(store)

    def test_mem_store_create_missing_key(self):
        store_meta = get_meta(NumberName)
        store_meta = StoreMeta(
            name=store_meta.name,
            attrs=tuple(
                dataclasses.replace(
                    f, write_transform=None, is_creatable=True, is_updatable=True
                )
                for f in store_meta.attrs
            ),
        )
        store = MemStore(store_meta)
        try:
            store.create({}) and self.assertTrue(False)
        except PersistyError:
            self.assertEqual(0, store.count())

    def test_mem_store_update_no_key(self):
        store_meta = get_meta(NumberName)
        store_meta = StoreMeta(
            name=store_meta.name,
            attrs=tuple(
                dataclasses.replace(
                    f, write_transform=None, is_creatable=True, is_updatable=True
                )
                for f in store_meta.attrs
            ),
        )
        store = MemStore(store_meta)
        try:
            store.update(dict(title="foobar")) and self.assertTrue(False)
        except PersistyError:
            self.assertEqual(0, store.count())
    """



from dataclasses import dataclass
from typing import Optional, Any, Iterator

from marshy.types import ExternalItemType

from persisty.key_config.attr_key_config import AttrKeyConfig
from persisty.attr.attr_type import AttrType
from persisty.key_config.key_config_abc import KeyConfigABC
from persisty.util.encrypt_at_rest import decrypt, encrypt


@dataclass(frozen=True)
class DynamodbKeyConfig(KeyConfigABC):
    pk_attr: AttrKeyConfig
    sk_attr: Optional[AttrKeyConfig] = None

    def to_key_str(self, item: ExternalItemType) -> str:
        if not self.sk_attr:
            key = self.pk_attr.get_value_for(item)
            key = str(key)
            return key
        return encrypt(
            [self.pk_attr.get_value_for(item), self.sk_attr.get_value_for(item)]
        )

    def from_key_str(self, key: Optional[str], output: Optional[Any] = None) -> Any:
        if output is None:
            output = {}
        if self.sk_attr:
            key = decrypt(key)
            self.pk_attr.set_value_for(output, key[0])
            self.sk_attr.set_value_for(output, key[1])
        else:
            if self.pk_attr.attr_type == AttrType.INT:
                key = int(key)
            elif self.pk_attr.attr_type == AttrType.FLOAT:
                key = float(key)
            self.pk_attr.set_value_for(output, key)
        return output

    def get_required_attrs(self) -> Iterator[str]:
        yield self.pk_attr
        yield self.sk_attr



from uuid import UUID

from persisty.factory.default_store_factory import DefaultStoreFactory
from persisty.impl.mem.mem_store import MemStore
from persisty.store_meta import get_meta
from persisty.stored import stored


@stored
class Message:
    id: UUID
    owner: str
    text: str


store = MemStore(get_meta(Message))
factory = DefaultStoreFactory(store)



from unittest import TestCase

from persisty.attr.generator.defaults import get_default_generator_for_create


class TestDefaults(TestCase):
    def test_str(self):
        creatable, generator = get_default_generator_for_create("id", str)
        self.assertFalse(creatable)
        self.assertEqual("1", generator.transform(None))
        self.assertEqual("2", generator.transform(None))

    def test_int(self):
        creatable, generator = get_default_generator_for_create("id", int)
        self.assertFalse(creatable)
        self.assertEqual(1, generator.transform(None))
        self.assertEqual(2, generator.transform(None))



import dataclasses
from unittest import TestCase
from uuid import UUID

from servey.security.authorization import Authorization

from persisty.factory.default_store_factory import DefaultStoreFactory
from persisty.factory.owned_store_factory import OwnedStoreFactory
from persisty.impl.mem.mem_store import MemStore
from persisty.store_meta import get_meta
from persisty.stored import stored


@stored
class Message:
    id: UUID
    owner: str
    text: str


class TestOwnedStoreFactory(TestCase):
    def test_getters(self):
        meta = get_meta(Message)
        store = MemStore(meta)
        meta = dataclasses.replace(meta, attrs=[
            meta.attrs[0],
            dataclasses.replace(meta.attrs[1], creatable=False, updatable=False),
            meta.attrs[2],
        ])
        factory = OwnedStoreFactory(DefaultStoreFactory(store), "owner")
        self.assertEqual(meta, factory.get_meta())
        subject_1 = Authorization("subject-1", frozenset(), None, None)
        subject_2 = Authorization("subject-2", frozenset(), None, None)
        subject_1_store = factory.create(subject_1)
        subject_2_store = factory.create(subject_2)
        msg_1 = subject_1_store.create(
            Message(id=UUID("f13dc535-9beb-488b-95d9-7f19f0d0a147"), text="Some test")
        )
        msg_2 = subject_2_store.create(
            Message(
                id=UUID("078709c0-0a7c-4a06-9599-b52511a6afa8"), text="Some other text"
            )
        )
        self.assertEqual("subject-1", msg_1.owner)
        self.assertEqual("subject-2", msg_2.owner)
        msg_2_updated = subject_1_store.update(msg_2)
        self.assertIsNone(msg_2_updated)
        msg_2_read = subject_2_store.read(msg_2.id)
        self.assertEqual(msg_2, msg_2_read)

    def test_create_actions(self):
        meta = get_meta(Message)
        store = MemStore(meta)
        factory = OwnedStoreFactory(DefaultStoreFactory(store), "owner")
        actions = factory.create_actions()
        action_names = {a.name for a in actions}
        expected_action_names = {
            "message_count",
            "message_create",
            "message_delete",
            "message_edit_batch",
            "message_read",
            "message_read_batch",
            "message_search",
            "message_update",
        }
        self.assertEqual(expected_action_names, action_names)
        self.assertEqual([], list(factory.create_routes()))



from uuid import UUID
from datetime import datetime, timezone

from marshy import dump

from persisty.attr.attr import Attr
from persisty.stored import stored


@stored(batch_size=10)
class NumberName:
    """Item linking a string representing a number with an integer value. Also has a uuid, and timestamps."""

    id: UUID
    title: str
    value: int = Attr(sortable=True)
    created_at: datetime
    updated_at: datetime


_num2words = {
    1: "One",
    2: "Two",
    3: "Three",
    4: "Four",
    5: "Five",
    6: "Six",
    7: "Seven",
    8: "Eight",
    9: "Nine",
    10: "Ten",
    11: "Eleven",
    12: "Twelve",
    13: "Thirteen",
    14: "Fourteen",
    15: "Fifteen",
    16: "Sixteen",
    17: "Seventeen",
    18: "Eighteen",
    19: "Nineteen",
    20: "Twenty",
    30: "Thirty",
    40: "Forty",
    50: "Fifty",
    60: "Sixty",
    70: "Seventy",
    80: "Eighty",
    90: "Ninety",
    0: "Zero",
}

NUMBER_NAMES = []
for n in range(1, 100):
    title = _num2words.get(n)
    if not title:
        title = _num2words[n - n % 10] + _num2words[n % 10].lower()
    NUMBER_NAMES.append(
        NumberName(
            id=UUID("00000000-0000-0000-0000-000000000" + (str(1000 + n)[1:])),
            title=title,
            value=n,
            created_at=datetime.fromtimestamp(0, tz=timezone.utc),
            updated_at=datetime.fromtimestamp(0, tz=timezone.utc),
        )
    )


NUMBER_NAMES_DICTS = [dump(n) for n in NUMBER_NAMES]



from enum import Enum
from unittest import TestCase

from schemey.schema import str_schema, int_schema

from persisty.attr.attr import Attr
from persisty.attr.attr_type import AttrType, attr_type
from persisty.util import UNDEFINED


class TestAttr(TestCase):
    def test_sanitize_type_str(self):
        attr = Attr("foo", AttrType.STR, str_schema())
        self.assertEqual("bar", attr.sanitize_type("bar"))

    def test_sanitize_type_int(self):
        attr = Attr("foo", AttrType.INT, int_schema())
        self.assertEqual(13, attr.sanitize_type(13))

    def test_sanitize_type_undefined(self):
        attr = Attr("foo", AttrType.STR, str_schema())
        self.assertIs(None, attr.sanitize_type(None))
        self.assertIs(UNDEFINED, attr.sanitize_type(UNDEFINED))

    def test_attr_type_enum(self):
        e = Enum("foobar", {"foo": 1, "bar": 2})
        self.assertEqual(AttrType.STR, attr_type(e))



from copy import deepcopy
from decimal import Decimal
from typing import Optional, Dict, Tuple, List
from dataclasses import dataclass, field

import boto3
import marshy
from boto3.dynamodb.conditions import Not as DynNot, ConditionBase, Key
from botocore.exceptions import ClientError
from marshy.types import ExternalItemType

from persisty.attr.attr import Attr
from persisty.attr.generator.attr_value_generator_abc import AttrValueGeneratorABC
from persisty.errors import PersistyError
from persisty.impl.dynamodb.partition_sort_index import PartitionSortIndex
from persisty.search_filter.and_filter import And
from persisty.search_filter.exclude_all import EXCLUDE_ALL
from persisty.search_filter.search_filter_abc import SearchFilterABC
from persisty.attr.attr_filter import AttrFilter, AttrFilterOp
from persisty.batch_edit import BatchEdit
from persisty.batch_edit_result import BatchEditResult
from persisty.result_set import ResultSet
from persisty.search_filter.include_all import INCLUDE_ALL
from persisty.search_order.search_order import SearchOrder
from persisty.store.store_abc import StoreABC, T
from persisty.store_meta import StoreMeta
from persisty.util import filter_none, get_logger
from persisty.util.undefined import UNDEFINED

logger = get_logger(__name__)


def catch_client_error(fn):
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except ClientError as e:
            raise PersistyError(e)

    return wrapper


@dataclass(frozen=True)
class DynamodbTableStore(StoreABC[T]):
    """Store backed by a dynamodb table. Does not do single table design or anything of that nature."""

    meta: StoreMeta
    table_name: str
    index: PartitionSortIndex
    global_secondary_indexes: Dict[str, PartitionSortIndex] = field(default_factory=dict)
    aws_profile_name: Optional[str] = None
    region_name: Optional[str] = None
    decimal_format: str = "%.9f"

    def get_meta(self) -> StoreMeta:
        return self.meta

    @catch_client_error
    def create(self, item: T) -> T:
        item = self._dump_create(item)
        self._dynamodb_table().put_item(
            Item=item,
            ConditionExpression=DynNot(self.index.to_condition_expression(item)),
        )
        loaded = self._load(item)
        return loaded

    @catch_client_error
    def read(self, key: str) -> Optional[T]:
        table = self._dynamodb_table()
        if not isinstance(key, str):
            key = str(key)
        key_dict = self.meta.key_config.to_key_dict(key)
        response = table.get_item(Key=key_dict)
        loaded = self._load(response.get("Item"))
        return loaded

    @catch_client_error
    def read_batch(self, keys: List[str]) -> List[Optional[T]]:
        assert len(keys) <= self.meta.batch_size
        key_config = self.meta.key_config
        resource = self._dynamodb_resource()

        kwargs = {
            "RequestItems": {
                self.table_name: {
                    "Keys": [key_config.to_key_dict(key) for key in set(keys)]
                }
            }
        }
        results_by_key = {}
        response = resource.batch_get_item(**kwargs)
        for item in response["Responses"][self.table_name]:
            loaded = self._load(item)
            key = key_config.to_key_str(loaded)
            results_by_key[key] = loaded
        assert not response.get(
            "UnprocessedKeys"
        )  # Batch size would have been greater than 16 Mb
        return [results_by_key.get(key) for key in keys]

    @catch_client_error
    def update(
        self, updates: T, search_filter: SearchFilterABC = INCLUDE_ALL
    ) -> Optional[T]:
        if search_filter is not INCLUDE_ALL:
            search_filter = search_filter.lock_attrs(self.meta.attrs)
            item = self.read(self.meta.key_config.to_key_str(updates))
            if not search_filter.match(item, self.meta.attrs):
                return None
        updates = self._dump_update(updates)
        key_dict = self.index.to_dict(updates)
        table = self._dynamodb_table()
        updates = {**updates}
        updates.pop(self.index.pk)
        if self.index.sk:
            updates.pop(self.index.sk)
        update = _build_update(updates)
        response = table.update_item(
            Key=key_dict,
            ConditionExpression=self.index.to_condition_expression(key_dict),
            UpdateExpression=update["str"],
            ExpressionAttributeNames=update["names"],
            ExpressionAttributeValues=update["values"],
            ReturnValues="ALL_NEW",
        )
        loaded = self._load(response.get("Attributes"))
        return loaded

    def _update(
        self,
        key: str,
        item: T,
        updates: T,
        search_filter: SearchFilterABC = INCLUDE_ALL,
    ) -> Optional[T]:
        return self.update(updates, search_filter)

    @catch_client_error
    def delete(self, key: str) -> bool:
        table = self._dynamodb_table()
        key_dict = self.meta.key_config.to_key_dict(key)
        response = table.delete_item(Key=key_dict, ReturnValues="ALL_OLD")
        attributes = response.get("Attributes")
        return bool(attributes)

    def _delete(self, key: str, item: T) -> bool:
        return self.delete(key)

    @catch_client_error
    def search(
        self,
        search_filter: SearchFilterABC = INCLUDE_ALL,
        search_order: Optional[SearchOrder] = None,
        page_key: Optional[str] = None,
        limit: Optional[int] = None,
    ) -> ResultSet[T]:
        if limit is None:
            limit = self.meta.batch_size
        assert limit <= self.meta.batch_size
        search_filter = search_filter.lock_attrs(self.meta.attrs)
        if search_order:
            search_order.validate_for_attrs(self.meta.attrs)
        if search_filter is EXCLUDE_ALL:
            return ResultSet([])
        index_name, condition, filter_expression, handled, descending = self.to_dynamodb_filter(
            search_filter
        )
        query_args = filter_none(
            {
                "KeyConditionExpression": condition,
                "IndexName": index_name,
                "Select": "SPECIFIC_ATTRIBUTES",
                "ProjectionExpression": ",".join(
                    a.name for a in self.meta.attrs if a.readable
                ),
                "FilterExpression": filter_expression,
                "Limit": limit,
            }
        )
        if descending:
            query_args["ScanIndexForward"] = False
        if page_key:
            query_args["ExclusiveStartKey"] = self.meta.key_config.to_key_dict(page_key)
        table = self._dynamodb_table()
        results = []
        while True:
            if condition:
                response = table.query(**query_args)
            else:
                response = table.scan(**query_args)
            items = [self._load(item) for item in response["Items"]]
            if not handled:
                items = [
                    item for item in items if search_filter.match(item, self.meta.attrs)
                ]
            results.extend(items)
            if len(results) >= limit:
                results = results[:limit]
                next_page_key = self.meta.key_config.to_key_str(results[-1])
                assert next_page_key != page_key  # Paranoid prevent infinite loop!
                return ResultSet(results, next_page_key)
            last_evaluated_key = response.get("LastEvaluatedKey")
            if not last_evaluated_key:
                return ResultSet(results)
            query_args["ExclusiveStartKey"] = last_evaluated_key

    @catch_client_error
    def count(self, search_filter: SearchFilterABC = INCLUDE_ALL) -> int:
        search_filter = search_filter.lock_attrs(self.meta.attrs)
        if search_filter is EXCLUDE_ALL:
            return 0
        index_name, condition, filter_expression, handled, _ = self.to_dynamodb_filter(
            search_filter
        )
        if not handled:
            logger.warning(f"search_filter_not_handled_by_dynamodb:{search_filter}")
            count = sum(1 for _ in self.search_all(search_filter))
            return count
        kwargs = filter_none(
            {
                "KeyConditionExpression": condition,
                "IndexName": index_name,
                "Select": "COUNT",
                "FilterExpression": filter_expression,
            }
        )
        table = self._dynamodb_table()
        count = 0
        while True:
            if condition:
                response = table.query(**kwargs)
            else:
                response = table.scan(**kwargs)
            count += response["Count"]  # Items
            last_evaluated_key = response.get("LastEvaluatedKey")
            kwargs["ExclusiveStartKey"] = last_evaluated_key
            if not last_evaluated_key:
                return count

    def _edit_batch(
        self, edits: List[BatchEdit], items_by_key: Dict[str, T]
    ) -> List[BatchEditResult]:
        assert len(edits) <= self.meta.batch_size
        results = []
        key_config = self.meta.key_config
        table = self._dynamodb_table()
        with table.batch_writer() as batch:
            for edit in edits:
                if edit.create_item:
                    item = self._dump_create(edit.create_item)
                    batch.put_item(Item=item)
                    results.append(BatchEditResult(edit, True))
                elif edit.update_item:
                    updates = edit.update_item
                    key = key_config.to_key_str(updates)
                    item = items_by_key[key]
                    to_put = {}
                    for attr in self.meta.attrs:
                        value = UNDEFINED
                        if attr.update_generator:
                            if attr.updatable:
                                value = attr.update_generator.transform(
                                    getattr(updates, attr.name)
                                )
                            else:
                                value = attr.update_generator.transform(UNDEFINED)
                        elif attr.updatable:
                            value = getattr(updates, attr.name)
                            if value is UNDEFINED:
                                value = getattr(item, attr.name)
                        if value is not UNDEFINED:
                            setattr(item, attr.name, value)
                            value = marshy.dump(value, attr.schema.python_type)
                            to_put[attr.name] = self._convert_to_decimals(value)
                    to_put = self._convert_to_decimals(to_put)
                    batch.put_item(Item=to_put)
                    edit.update_item = deepcopy(item)  # In case of multi put
                    results.append(BatchEditResult(edit, True))
                else:
                    key = key_config.to_key_dict(edit.delete_key)
                    batch.delete_item(Key=key)
                    results.append(BatchEditResult(edit, True))
        return results

    def _load(self, item) -> T:
        if item is None:
            return None
        item = self._convert_from_decimals(item)
        kwargs = {}
        for attr in self.meta.attrs:
            value = item.get(attr.name, UNDEFINED)
            if attr.readable and value is not UNDEFINED:
                # noinspection PyTypeChecker
                kwargs[attr.name] = marshy.load(attr.schema.python_type, value)
        result = self.meta.get_read_dataclass()(**kwargs)
        return result

    def _convert_from_decimals(self, item):
        if isinstance(item, dict):
            return {k: self._convert_from_decimals(v) for k, v in item.items()}
        elif isinstance(item, list):
            return [self._convert_from_decimals(i) for i in item]
        elif isinstance(item, Decimal):
            int_val = int(item)
            if int_val == item:
                return int_val
            else:
                return float(item)
        else:
            return item

    def _dump_create(self, to_create: T):
        result = {}
        for attr in self.meta.attrs:
            self._dump_attr(
                to_create, attr, attr.creatable, attr.create_generator, result
            )
        return result

    def _dump_update(self, to_update: T):
        result = {}
        for attr in self.meta.attrs:
            self._dump_attr(
                to_update, attr, attr.updatable, attr.update_generator, result
            )
        return result

    def _dump_attr(
        self,
        item: T,
        attr: Attr,
        accepts_input: bool,
        generator: Optional[AttrValueGeneratorABC],
        target: ExternalItemType,
    ):
        value = UNDEFINED
        if accepts_input:
            value = getattr(item, attr.name, UNDEFINED)
        if generator:
            value = generator.transform(value)
        if value is not UNDEFINED:
            value = marshy.dump(value, attr.schema.python_type)
            target[attr.name] = self._convert_to_decimals(value)

    def _convert_to_decimals(self, item):
        if isinstance(item, dict):
            return {k: self._convert_to_decimals(v) for k, v in item.items()}
        elif isinstance(item, list):
            return [self._convert_to_decimals(i) for i in item]
        elif isinstance(item, float):
            int_val = int(item)
            if int_val == item:
                return int_val
            else:
                return Decimal(self.decimal_format % item)
        else:
            return item

    def to_dynamodb_filter(
        self,
        search_filter: SearchFilterABC,
    ) -> Tuple[Optional[str], Optional[ConditionBase], Optional[ConditionBase], bool, bool]:
        eq_filters = _get_top_level_eq_filters(search_filter)
        if not eq_filters:
            filter_expression, handled = search_filter.build_filter_expression(
                self.meta.attrs
            )
            return None, None, filter_expression, handled, False
        index_name, index = self.get_index_for_eq_filters(eq_filters)
        filter_expression = None
        handled = False
        if index:
            index_condition, search_filter, index_handled, reverse = _separate_index_filters(
                index, eq_filters
            )
            if search_filter:
                filter_expression, handled = search_filter.build_filter_expression(
                    self.meta.attrs
                )
            else:
                handled = True
            return (
                index_name,
                index_condition,
                filter_expression,
                handled and index_handled,
                reverse
            )
        if search_filter:
            filter_expression, handled = search_filter.build_filter_expression(
                self.meta.attrs
            )
        return None, None, filter_expression, handled, False

    def get_index_for_eq_filters(
        self, eq_filters: List[AttrFilter]
    ) -> Tuple[Optional[str], Optional[PartitionSortIndex]]:
        attr_names = {f.name for f in eq_filters}
        if self.index.pk in attr_names:
            return None, self.index
        for name, index in self.global_secondary_indexes.items():
            if index.pk in attr_names:
                return name, index
        return None, None

    def _dynamodb_table(self):
        if hasattr(self, "_table"):
            return self._table
        resource = self._dynamodb_resource()
        table = resource.Table(self.table_name)
        object.__setattr__(self, "_table", table)
        return table

    def _dynamodb_resource(self):
        if hasattr(self, "_resource"):
            return self._resource
        kwargs = filter_none(
            dict(profile_name=self.aws_profile_name, region_name=self.region_name)
        )
        session = boto3.Session(**kwargs)
        resource = session.resource("dynamodb")
        object.__setattr__(self, "_resource", resource)
        return resource


def _build_update(updates: dict):
    update_str = "set "
    update_list = []
    names = {}
    values = {}
    for k, v in updates.items():
        update_list.append(f"#n_{k} = :v_{k}")
        names[f"#n_{k}"] = k
        values[f":v_{k}"] = v
    update_str += ", ".join(update_list)
    return {"str": update_str, "names": names, "values": values}


# noinspection PyTypeChecker
def _get_top_level_eq_filters(search_filter: SearchFilterABC) -> List[AttrFilter]:
    if _is_eq_filter(search_filter):
        return [search_filter]
    elif isinstance(search_filter, And):
        return [f for f in search_filter.search_filters if _is_eq_filter(f)]
    return []


def _is_eq_filter(search_filter: SearchFilterABC) -> bool:
    return isinstance(search_filter, AttrFilter) and search_filter.op == AttrFilterOp.eq


def _separate_index_filters(
    index: PartitionSortIndex, eq_filters: List[AttrFilter]
) -> Tuple[ConditionBase, SearchFilterABC, bool, bool]:
    index_filters = [f for f in eq_filters if f.name == index.pk]
    value = index_filters[0].value
    if value.__class__ not in (str, int, float, bool):
        value = marshy.dump(value)
    condition = Key(index.pk).eq(value)
    non_index_filters = tuple(f for f in eq_filters if f.name != index.pk)
    non_index_filter = None
    if non_index_filters:
        non_index_filter = And(non_index_filters)
    handled = len(index_filters) == 1
    return condition, non_index_filter, handled, index.descending



import setuptools

from persisty.__version__ import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

extras_require = {
    "sql": ["SQLAlchemy~=1.4"],
    "server": ["servey[server]~=2.7"],
    "serverless": ["servey[serverless]~=2.7"]
}

setuptools.setup(
    name="persisty",
    version=__version__,
    author="Tim O'Farrell",
    author_email="tofarr@gmail.com",
    description="A better persistence layer for python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tofarr/lambsync",
    packages=setuptools.find_packages(exclude=("tests",)),
    install_requires=[
        "marshy~=3.0",
        "pyaes~=1.6",
        "schemey~=5.7",
        "servey~=2.7",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
